/* R[h͂rihr sR[h͂bqke */
/* $Id: Predict.cpp 480 2010-01-15 07:02:44Z umezawa $ */

#include "StdAfx.h"
#include "utvideo.h"
#include "Predict.h"
#include "TunedFunc.h"

inline BYTE median(BYTE a, BYTE b, BYTE c)
{
	return max(min(max(a,b),c),min(a,b));
}

void PredictMedian(BYTE *pDst, const BYTE *pSrcBegin, const BYTE *pSrcEnd, DWORD dwStride)
{
	if (IS_ALIGNED(pDst, 16) && IS_ALIGNED(pSrcBegin, 16) && IS_ALIGNED(dwStride, 16))
		tfn.pfnPredictMedian_align16(pDst, pSrcBegin, pSrcEnd, dwStride);
	else
		cpp_PredictMedian(pDst, pSrcBegin, pSrcEnd, dwStride);
}

void PredictMedianAndCount(BYTE *pDst, const BYTE *pSrcBegin, const BYTE *pSrcEnd, DWORD dwStride, DWORD *pCountTable)
{
	if (IS_ALIGNED(pDst, 16) && IS_ALIGNED(pSrcBegin, 16) && IS_ALIGNED(dwStride, 16))
		tfn.pfnPredictMedianAndCount_align16(pDst, pSrcBegin, pSrcEnd, dwStride, pCountTable);
	else
		tfn.pfnPredictMedianAndCount_align1(pDst, pSrcBegin, pSrcEnd, dwStride, pCountTable);
}

void cpp_PredictMedian(BYTE *pDst, const BYTE *pSrcBegin, const BYTE *pSrcEnd, DWORD dwStride)
{
	const BYTE *p = pSrcBegin;
	BYTE *q = pDst;

	// ŏ̃C̍ŏ̃sNZ 0x80 \ĂB
	*q++ = *p++ - 0x80;

	// ŏ̃C̎c̃sNZ predict left ƓB
	for (; p < pSrcBegin + dwStride; p++, q++)
	{
		*q = *p - *(p - 1);
	}

	// 1 CȂꍇ͂ŏIB
	if (p == pSrcEnd)
		return;

	// ̃C̍ŏ̃sNZ predict aboveB
	// ĂƃAZuɏ኱ȒPɂȂB
	*q = *p - *(p - dwStride);
	p++;
	q++;

	// c̃sNZ predict median ̖{
	for (; p < pSrcEnd; p++, q++)
	{
		*q = *p - median(*(p - dwStride), *(p - 1), *(p - dwStride) + *(p - 1) - *(p - 1 - dwStride));
	}
}

void cpp_PredictMedianAndCount(BYTE *pDst, const BYTE *pSrcBegin, const BYTE *pSrcEnd, DWORD dwStride, DWORD *pCountTable)
{
	const BYTE *p = pSrcBegin;
	BYTE *q = pDst;

	// ŏ̃C̍ŏ̃sNZ 0x80 \ĂB
	*q = *p - 0x80;
	pCountTable[*q]++;
	p++;
	q++;

	// ŏ̃C̎c̃sNZ predict left ƓB
	for (; p < pSrcBegin + dwStride; p++, q++)
	{
		*q = *p - *(p - 1);
		pCountTable[*q]++;
	}

	// 1 CȂꍇ͂ŏIB
	if (p == pSrcEnd)
		return;

	// ̃C̍ŏ̃sNZ predict aboveB
	// ĂƃAZuɏ኱ȒPɂȂB
	*q = *p - *(p - dwStride);
	pCountTable[*q]++;
	p++;
	q++;

	// c̃sNZ predict median ̖{
	for (; p < pSrcEnd; p++, q++)
	{
		*q = *p - median(*(p - dwStride), *(p - 1), *(p - dwStride) + *(p - 1) - *(p - 1 - dwStride));
		pCountTable[*q]++;
	}
}

void cpp_PredictLeftAndCount(BYTE *pDst, const BYTE *pSrcBegin, const BYTE *pSrcEnd, DWORD *pCountTable)
{
	const BYTE *p = pSrcBegin;
	BYTE *q = pDst;

	// ŏ̃C̍ŏ̃sNZ 0x80 \ĂB
	*q = *p - 0x80;
	pCountTable[*q]++;
	p++;
	q++;

	// c̃sNZ predict left ̖{
	for (; p < pSrcEnd; p++, q++)
	{
		*q = *p - *(p-1);
		pCountTable[*q]++;
	}
}

void cpp_RestoreMedian(BYTE *pDst, const BYTE *pSrcBegin, const BYTE *pSrcEnd, DWORD dwStride)
{
	const BYTE *p = pSrcBegin;
	BYTE *q = pDst;

	*q++ = *p++ + 0x80;

	for (; p < pSrcBegin + dwStride; p++, q++)
	{
		*q = *p + *(q - 1);
	}

	if (p == pSrcEnd)
		return;

	*q = *p + *(q - dwStride);
	p++;
	q++;

	for (; p < pSrcEnd; p++, q++)
	{
		*q = *p + median(*(q - dwStride), *(q - 1), *(q - dwStride) + *(q - 1) - *(q - 1 - dwStride));
	}
}
